from torchhd.datasets.beijing_air_quality import BeijingAirQuality
from torchhd.datasets.isolet import ISOLET
from torchhd.datasets.european_languages import EuropeanLanguages
from torchhd.datasets.ucihar import UCIHAR
from torchhd.datasets.airfoil_self_noise import AirfoilSelfNoise
from torchhd.datasets.emg_hand_gestures import EMGHandGestures
from torchhd.datasets.pamap import PAMAP
from torchhd.datasets.ccpp import CyclePowerPlant

__all__ = [
    "BeijingAirQuality",
    "ISOLET",
    "EuropeanLanguages",
    "UCIHAR",
    "AirfoilSelfNoise",
    "EMGHandGestures",
    "PAMAP",
    "CyclePowerPlant",
]
